/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * END COPYRIGHT BLOCK **/
/*
 * Description (listOldSrvs.c)
 *
 *	This module contains routines used by administration CGI
 *	programs to retrieve the server instances from an old server root
 *
 */
#ifdef XP_UNIX
#include <unistd.h>
#endif
#include <stdio.h>
#include <string.h>
#include "prio.h"
#include "plstr.h"
#include "prmem.h"
#include "prprf.h"
#include "libadminutil/resource.h"
#include "libadminutil/distadm.h"
#include "libadminutil/admutil.h"
#include "libadminutil/resource.h"
#   define SYS_DIR PRDir
#   define SYS_DIRENT PRDirEntry
#   define dir_open PR_OpenDir
#   define dir_read(d) PR_ReadDir(d, PR_SKIP_BOTH)
#   define dir_close PR_CloseDir
#   define dir_name(de) (de->name)


#define RESOURCE_FILE "listOldSrvs"

#define resource_key(a,b)   a b

#define DBT_NO_METHOD 			    resource_key(RESOURCE_FILE, "1")
#define DBT_UNKNOWN_METHOD 		    resource_key(RESOURCE_FILE, "2")
#define DBT_NO_QUERY_STRING 	    resource_key(RESOURCE_FILE, "3")
#define DBT_ERROR_OPEN_FILE 		resource_key(RESOURCE_FILE, "4")
#define DBT_ERROR_READ_FILE 		resource_key(RESOURCE_FILE, "5") 
#define DBT_ERROR_OPEN_DIR 	        resource_key(RESOURCE_FILE, "6") 
#define DBT_DBT_NO_OLD_SERVER_ROOT 	resource_key(RESOURCE_FILE, "7") 
#define DBT_NO_SERVER_FOUND 		resource_key(RESOURCE_FILE, "8") 


char           **installed = NULL;
int            num_installed = 0;
Resource       *i18nResource = NULL;
char           *acceptLanguage = "en";


int
readServersList(char* serverRoot, FILE *fstream, char *fileName, char* errorInfo, size_t errorSize)
{
  char		line[1024];
  int		ln;
  int		stop;
  int		t;
  const char    *errorMsg = NULL;

 
  stop=0; ln=0;
  while(!stop)  {
    char *d;
    char *v;

    ++ln;
    switch( (t = admutil_getline(fstream, sizeof(line), ln, line))) {
    case -1:
      fclose(fstream);
      if (i18nResource && 
	  (errorMsg = res_getstring(i18nResource, 
				    DBT_ERROR_READ_FILE,
				    acceptLanguage)))
	PR_snprintf(errorInfo, errorSize, errorMsg, fileName);
      else PR_snprintf(errorInfo, errorSize, "Error reading file: %s", fileName);
      return 0;
    case 1:
      fclose(fstream);
      stop=1;
      break;
 
    default:
      d = line;
      v = PL_strchr(line, ':');
      if (v == NULL) break;
      *v++ = '\0';
      if(!installed)  {
	num_installed++;
	installed = (char **) PR_Malloc((num_installed+1)*sizeof(char *));
      }  
      else  {
	int	i;

	for ( i = 0 ; i < num_installed ; ++i ) {
	  if ( PL_strcmp( d, installed[i] ) == 0 ) {
	    break;	/* already read this one */
	  }
	}
	if ( i < num_installed ) {
	  break;	/* already read it */
	}
	num_installed++;
	installed = (char **) PR_Realloc(installed,(num_installed+1)*sizeof(char *));
      }
      installed[num_installed-1] = PL_strdup(d);
      installed[num_installed] = NULL;
      break;
    }
  }
  return 1;
}

/* given the name of a directory and the name of a file/directory in that parent
   return true if the given file/directory exists */
static int
file_or_dir_exists(const char *parent, const char *name)
{
    SYS_DIR       *dp;
    SYS_DIRENT    *d;
    int ret = 0; /* default to false */

    dp = dir_open(parent);
    /* loop through all directory entries until we find the one that matches */
    while (dp && name && !ret && (d = dir_read(dp))) {
	const char *entname = dir_name(d);
	/* exclude "." and ".." from scanning */
	if (strcmp(entname, ".") && strcmp(entname, "..")) {
	    /* set ret to true if the name matches, which terminates the while loop */
	    ret = !strcmp(name, entname);
	}
    }
    if (dp) {
	dir_close(dp);
    }

    return ret;
}

char** 
get_server_list(char* oldServerRoot, char* errorInfo, size_t errorSize) {

  PRDir 	*sr;
  PRDirEntry 	*dirname;
  FILE          *fstream;
  int 		result, num = 0, i, match;
  char  	**servers = NULL;
  char          *fileName;
  const char    *errMsg = NULL;

  if (!oldServerRoot) return 0;

  /* since oldServerRoot is passed in from a (potentially malicious) user, we cannot
	 assume it will fit into a buffer of MAX_PATH/PATH_MAX size e.g. what if the
	 value were something like /etc/../etc/../etc/../ ....... ../etc/passwd which
	 was exactly 255 bytes long?  If we just did an PR_snprintf of this string into
	 a buffer of size 256, the final fileName value would end up being /etc/passwd */
  /* first, construct the base name from the given server root */
  fileName = PR_smprintf(oldServerRoot); /* essentially, just a strdup */
  if (!fileName) {
	  goto errorreturn;
  }
  /* next, see if the path has an "admin-serv" sub directory - also tests if fileName is NULL */
  if (file_or_dir_exists(fileName, "admin-serv")) {
	  /* ok, admin-serv subdir exists - append that to fileName and look for config */
	  fileName = PR_sprintf_append(fileName, "/admin-serv");
	  /* next, see if the path has an "config" sub directory - also tests if fileName is NULL */
	  if (!fileName || !file_or_dir_exists(fileName, "config")) {
		  goto errorreturn;
	  }
	  /* ok, config subdir exists - append that to fileName and look for servers.lst */
	  fileName = PR_sprintf_append(fileName, "/config");
	  /* next, see if the path has an "servers.lst" file - also tests if fileName is NULL */
	  if (!fileName || !file_or_dir_exists(fileName, "servers.lst")) {
		  goto errorreturn;
	  }
	  /* ok, servers.lst exists, append it */
  } else if (file_or_dir_exists(fileName, "admserv")) {
	/* ok, admserv subdir exists - append that to fileName and look for servers.lst */
	fileName = PR_sprintf_append(fileName, "/admserv");
	/* next, see if the path has an "servers.lst" file - also tests if fileName is NULL */
	if (!fileName || !file_or_dir_exists(fileName, "servers.lst")) {
		goto errorreturn;
	}
  } else {
	goto errorreturn;
  }
  /* ok, servers.lst exists, append it */
  fileName = PR_sprintf_append(fileName, "/servers.lst");

  if(!fileName || !(fstream = fopen(fileName, "r"))) {
	  goto errorreturn;
  }
  
  if (!(result = readServersList(oldServerRoot, fstream, fileName, errorInfo, errorSize))) {
	PR_smprintf_free(fileName);
    return NULL;
  }

  sr = PR_OpenDir(oldServerRoot);
  if (!sr) {
    if (i18nResource &&
	(errMsg = res_getstring(i18nResource, 
				DBT_ERROR_OPEN_DIR,
				acceptLanguage)))
      PR_snprintf(errorInfo, errorSize, errMsg, oldServerRoot);
    else PR_snprintf(errorInfo, errorSize, "Error open directory: %s", oldServerRoot);
    return NULL;
  }


  while (dirname = PR_ReadDir(sr, PR_SKIP_BOTH)) {
    match = 0;
    if (!PL_strcmp(dirname->name, "admin-serv")) match = 1;
    if (!match && num_installed) {
      for (i=0; i < num_installed; i++) {
	if (!PL_strncmp(dirname->name, installed[i], PL_strlen(installed[i])))  {
	  if (dirname->name[PL_strlen(installed[i])] == '-') match = 1;
	}
      }
    }
    if (match) {
      if (!servers) servers = (char **) PR_Malloc((num+2)*sizeof(char*));
      else servers = (char **) PR_Realloc(servers, (num+2)*sizeof(char*));
      servers[num++] = PL_strdup(dirname->name);
      servers[num] = NULL;
    }
  }
  PR_CloseDir(sr);
  return servers;

errorreturn:
  if (fileName) {
	  if (i18nResource &&
		  (errMsg = res_getstring(i18nResource, 
								  DBT_ERROR_OPEN_FILE,
								  acceptLanguage))) {
		  PR_snprintf(errorInfo, errorSize, errMsg, fileName);
	  } else {
		  PR_snprintf(errorInfo, errorSize, "Error open file: %s", fileName);
	  }
	  PR_smprintf_free(fileName);
  }

  return NULL;
}



int main(int argc, char *argv[])
{
  int            _ai = ADMUTIL_Init();
  char           *method;
  int            rv, cnt, i, j, x, errorCode;
  char           *url;
  int            err;
  int            *errp = &err;
  char           *oldSR = 0;
  char           **server_list = 0;
  char           *operation = 0;
  char           *qs = 0;
  char           *nameptr, *valptr, *val;
  char           error_info[128];
  int            setFlag = 0, getFlag = 0, forceSetFlag = 0;
  char           *execPath, *tmpptr;
  char           *lang = getenv("HTTP_ACCEPT_LANGUAGE");
  const char     *errMsg = NULL;
#if 0
  int            waitforever = 1;

  while (waitforever);
#endif

  i18nResource = res_init_resource("bin/admin/admin/bin/property",
				   RESOURCE_FILE);
  if (lang) acceptLanguage = strdup(lang);


  memset((void *)errp, 0, sizeof(int));
  method = getenv("REQUEST_METHOD");

  /* Get new port number */
  if (!method || !*method) {
    /* non UI CGI */
    if (i18nResource &&
	(errMsg = res_getstring(i18nResource, 
				DBT_NO_METHOD,
				acceptLanguage)))
      rpt_err(SYSTEM_ERROR, errMsg, NULL, NULL);
    else rpt_err(SYSTEM_ERROR, "No method is specified", NULL, NULL);
  }
  else if (!PL_strcmp(method, "GET")) {
    /* UI CGI - either upgrade or normal admin CGI */
    qs = getenv("QUERY_STRING");
	
    if (!qs || !*qs) {
      if (i18nResource &&
	  (errMsg = res_getstring(i18nResource, 
				  DBT_NO_QUERY_STRING,
				  acceptLanguage)))
	rpt_err(INCORRECT_USAGE, errMsg, NULL, NULL);
      else rpt_err(INCORRECT_USAGE, "NO QUERY_STRING DATA", NULL, NULL);
    }
    else {
      get_begin(qs);
    }
  }
  else if (!PL_strcmp(method, "POST")) {
    post_begin(stdin);
  }
  else {
    if (i18nResource &&
	(errMsg = res_getstring(i18nResource, 
				DBT_UNKNOWN_METHOD, 
				acceptLanguage)))
      PR_snprintf(error_info, sizeof(error_info), errMsg, method);
    else PR_snprintf(error_info, sizeof(error_info), "Unknown Method (%s)", method);
    rpt_err(INCORRECT_USAGE, error_info, NULL, NULL);
  }

  oldSR = get_cgi_var("oldServerRoot", NULL, NULL);

  if (!oldSR) {
    if (i18nResource &&
	(errMsg = res_getstring(i18nResource, 
				DBT_DBT_NO_OLD_SERVER_ROOT,
				acceptLanguage)))
       rpt_err(INCORRECT_USAGE, errMsg, NULL, NULL);
    else rpt_err(INCORRECT_USAGE, "No old server root specified", NULL, NULL);
  }

  server_list = get_server_list(oldSR, error_info, sizeof(error_info));

  if (server_list) {
    rpt_success(NULL);
    x = 0;
    fprintf(stdout, "SR:");
    while (server_list[x]) fprintf(stdout, " %s", server_list[x++]);
    fprintf(stdout, "\n");
  }
  else {
    if (i18nResource &&
	(errMsg = res_getstring(i18nResource, 
				DBT_NO_SERVER_FOUND,
				acceptLanguage)))
      rpt_warning(WARNING, errMsg, NULL, NULL);
    else
      rpt_warning(WARNING, "No server found", NULL, NULL);
  }
  if (i18nResource) res_destroy_resource(i18nResource);
}
