/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * END COPYRIGHT BLOCK **/

extern "C" {
#include <stdio.h>		/* printf, file I/O */
#include <string.h>		/* strlen */
#include <strings.h>
#include <stdlib.h>		/* memset, rand stuff */
#include <time.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>

#include "ldap.h"
#include "libadminutil/admutil.h"
#include "libadminutil/psetc.h"

#if defined(AIX)
extern int seteuid(uid_t);
extern int setegid(gid_t);
#endif

}

#include "global.h"
#include "code.h"
#include "ux-util.h"
#include "ldapu.h"
#include "setupldap.h"
#include "uninstall.h"
#include "../common/include/acidef.h"

static int set_locate_admserv_aci(Ldap *ldap);
static int repair_group_expansion_aci(Ldap *ldap);
static void replaceTokensInFile(const char *sourcefile, const char *destfile, const char * const *ary);

#ifdef HPUX
  #define SHLIB_EXT "sl"
#else
  #define SHLIB_EXT "so"
#endif

#define EPOCH_32BIT_END_TIME  "20380118031417" /* <correct end time> minus 1 day */

const char *DEFAULT_SYSUSER = "root";
const char *DEFAULT_TASKCONF = "admin-serv/ldif/tasks.ldif";
const char *DEFAULT_GLOBALCONF = "admin-serv/ldif/global.ldif";

const char *CHOWN_CMD = "chown -R %s %s >/dev/null 2>&1";
const char *CHGRP_CMD = "chgrp -R %s %s >/dev/null 2>&1";
const char *SHARED_CHGRP_CMD = "chgrp -R %s %s/shared >/dev/null 2>&1";
const char *SSCONF_CHGRP_CMD = "chgrp -R %s %s/shared/config/ssusers.conf >/dev/null 2>&1";
const char *SECALIAS_CHGRP_CMD = "chgrp -R %s %s/alias >/dev/null 2>&1";

const char *DEFAULT_ADMIN_CONFIG_ACI = "(targetattr=*)(version 3.0; acl \"Enable delegated admin to access configuration\"; allow (read,search) groupdn=\"ldap:///%s\";)";

InstallMode installMode = Interactive;
InstallInfo *installInfo = NULL;
InstallInfo *adminInstallInfo = NULL;
InstallInfo *dsInfo = NULL;
char *infoFile = NULL;
char *logFile = NULL;
InstallLog    *installLog = NULL;
Bool reConfig = False;

char *
system_errmsg(void)
{
    return strerror(errno);
}

void
fatal_error(const char *format, const char *arg1, const char *arg2)
{
   if (installMode == Interactive)
   {
      printf(format,arg1, arg2);
      printf("\n");
   }
   else
   {
      printf("ERROR: Administration Server configuration failed.\n");
      installLog->logMessage(FATAL, "Admin", format, arg1, arg2);
   }
   exit(1);
}

/*
 * The following escape LDAP URL is from ns/netsite/ldap/clients/dsgw/htmlout.c
 */

#define HREF_CHAR_ACCEPTABLE( c )       (( c >= '-' && c <= '9' ) ||    \
                                         ( c >= '@' && c <= 'Z' ) ||    \
                                         ( c == '_' ) ||                \
                                         ( c >= 'a' && c <= 'z' ))
void
escape_ldap_uri(char *s1, char *s2)
{
   unsigned char *q;
   char *p, *hexdig = "0123456789ABCDEF";

   p = s1 + strlen(s1);
   for (q = (unsigned char *) s2; *q != '\0'; ++q)
   {
      if (HREF_CHAR_ACCEPTABLE(*q))
      {
	 *p++ = *q;
      }
      else
      {
	 *p++ = '%';
	 *p++ = hexdig[*q >> 4];
	 *p++ = hexdig[*q & 0x0F];
      }
   }

   *p = '\0';
}

/* ----------------------- make a new admin server ------------------------ */

char *
make_shexp(char *source)
{
   char *answer;
   int hascomma, x, y;

   if (!source)
      return source;

   hascomma = (strchr(source, ',')) ? 1 : 0;

   /* + 2 = 1 for trailing \0 and one put in by Aruna for NT */
   answer = (char *) malloc(strlen(source) + 2);

   for (x = 0, y = 0; source[x]; x++)
   {
      if (source[x] == ',')
	 answer[y++] = '|';
      else if (source[x] != ' ')
	 answer[y++] = source[x];
      answer[y] = '\0';
   }
   if (!hascomma)
      return (answer);
   else
   {
      char *newanswer = (char *) malloc(strlen(answer) + 3);
      sprintf(newanswer, "(%s)", answer);
      return (newanswer);
   }
}

/* Be tricky and check the syntax of the URL at the same time it
   escapes the base DN.  Return escaped string if everything when according to
   plan, else NULL */
static char *
escapeLDAPurl(char *url)
{
   char *result = (char *) malloc(strlen(url) * 3 + 1);	/* overkill */
   int i;

   for (i = 0; url[i] && url[i] != ':'; ++i)
   {
      result[i] = url[i];
   }
   result[i] = '\0';
   if (strcmp(result, "ldap") != 0 && strcmp(result, "ldaps") != 0)
   {
      return NULL;
   }
   result[i++] = ':';
   if ((result[i] = url[i]) != '/')
      return NULL;
   ++i;
   if ((result[i] = url[i]) != '/')
      return NULL;
   ++i;
   while ((result[i] = url[i]) && (result[i] != '/'))
   {
      ++i;
   }
   if (!result[i])
      return NULL;
   result[++i] = '\0';
   escape_ldap_uri(result + i, url + i);

   return result;
}

static char *get_value(char * file, char * attr) {
   char cmd[1024];
   char buf[1024];
   FILE *fp;

   sprintf(cmd, "%s -V", file);
   fp = popen(cmd, "r");

   if (fp != NULL) {
      char *x, *t;
      while ((fgets(buf,  sizeof(buf), fp)) > 0) {
         x = (char *)strstr(buf, attr);
         if (x) {
            t = (char *)strtok(buf, "\"");
            if (t) {
               t = (char *)strtok(NULL, "\"");
               if (t) {
                  pclose(fp);
                  return (char *)(strdup(t));
               }
            }
         }
      }
   }
   pclose(fp);
   return NULL;
}

static NSString
findApacheRoot(const char *dir)
{
   char path[1024];
   struct stat st;
   char *v;

   snprintf(path, sizeof(path), "%s/httpd.worker", dir);
   if (stat(path, &st) != 0) {
      snprintf(path, sizeof(path), "%s/httpd", dir);
      if (stat(path, &st) != 0) {
         return NULL;
      }
   }

   v = get_value(path, "HTTPD_ROOT");
   if (v) {
	   sprintf(path, "%s/modules", v);
	   if (stat(path, &st) != 0) {
		   return NULL;
	   }
   }

   return v;
}

static char *get_version(char * file) {
   char cmd[1024];
   char buf[1024];
   FILE *fp;
   const char *versstr = "Server version: Apache/";

   sprintf(cmd, "%s -v", file);
   fp = popen(cmd, "r");

   if (fp != NULL) {
      char *x, *t;
      while ((fgets(buf,  sizeof(buf), fp)) > 0) {
         x = (char *)strstr(buf, versstr);
         if (x) {
            t = x + strlen(versstr);
            if (x = strchr(buf, '\n')) {
                *x = '\0';
            }
            pclose(fp);
            return (char *)(strdup(t));
         }
      }
   }
   pclose(fp);
   return NULL;
}

static NSString
findApacheVersion(const char *dir)
{
   char path[1024];
   struct stat st;
   char *v;

   snprintf(path, sizeof(path), "%s/httpd.worker", dir);
   if (stat(path, &st) != 0) {
      snprintf(path, sizeof(path), "%s/httpd", dir);
      if (stat(path, &st) != 0) {
         return NULL;
      }
   }

   v = get_version(path);

   return v;
}

void
configTasks(const Ldap *ldap, const char *sroot, const char *sieDN)
{
   char tmp[BIG_BUF];
   LdapErrorCode err;

   printf("Configuring Administration Tasks in Directory Server...\n");
   installLog->logMessage(INFO, "Admin", "Configuring Administration Tasks in Directory Server");

   sprintf(tmp, "%s/%s", sroot, DEFAULT_TASKCONF);

   err = setupReplaceLdifEntries(ldap, sieDN, tmp, sieDN);

   if (err != OKAY)
   {
      printf("Cannot write task configuration. Error (%ld).\n", err);
   }
   return;
}

void
configGlobalParameters(const Ldap *ldap, const char *sroot, const char *subTreeDN)
{
   char tmp[BIG_BUF];
   LdapErrorCode err;

   printf("Configuring Global Parameters in Directory Server...\n");
   installLog->logMessage(INFO, "Admin", "Configuring global parameters...");

   sprintf(tmp, "%s/%s", sroot, DEFAULT_GLOBALCONF);

   err = setupUpdateLdifEntries(ldap, subTreeDN, tmp, NULL);

   if (err != OKAY)
   {
      printf("Cannot write Admin global parameters. Error (%ld).\n", err);
   }
   return;
}


/*
 * Cleanup prevoius installation after upgrading from 6.0x
 */
int
cleanup_after_upgrade(const char *sroot) 
{
   struct stat fi;
   char buf[BIG_BUF];

   // Move admin-serv/config/secmod.db to the alias directory;
   // Starting with as 6.1 secmod.db is in the alias directory.
   sprintf(buf, "%s/admin-serv/config/secmod.db", sroot);
   if (stat(buf, &fi) == 0) {
       char to[BIG_BUF];
       sprintf(to, "%s/alias/secmod.db", sroot);
       if (link(buf, to) == 0) { 
           unlink(buf);
       }
   }

   // Remove admin-serv/config/libnssckbi.so;
   // Starting with as 6.1 secmod.db is in the alias directory.
   sprintf(buf, "%s/admin-serv/config/libnssckbi.%s", sroot, SHLIB_EXT);
   if (stat(buf, &fi) == 0) {
       unlink(buf);
   }

   // Remove admin-serv/config/res directory. Not packaged starting with as61.
   sprintf(buf, "%s/admin-serv/config/res", sroot);
   if (stat(buf, &fi) == 0) {
       char cmd[BIG_BUF];
       sprintf(cmd, "rm -fr %s", buf);
       system(cmd);
   }

   // Remove adminsdk 60
   sprintf(buf, "%s/lib/libadminutil60.%s", sroot, SHLIB_EXT);
   if (stat(buf, &fi) == 0) {
       unlink(buf);
   }
   sprintf(buf, "%s/lib/libadmsslutil60.%s", sroot, SHLIB_EXT);
   if (stat(buf, &fi) == 0) {
       unlink(buf);
   }

   // Remove admserv jars from the previous installation
   sprintf(buf, "%s/java/jars/admserv60.jar", sroot);
   if (stat(buf, &fi) == 0) {
       char cmd[BIG_BUF];
       sprintf(cmd, "rm %s/java/jars/admserv60*", sroot);
       system(cmd);
   }

   return 0;
}


void
start_admserv(const char *sroot)
{
   char pid[BIG_BUF];
   char startProgram[BIG_BUF];
   char startLog[SML_BUF];
   struct stat fi;
   int ret;

   sprintf(pid, "%s/admin-serv/logs/pid", sroot);

   tmpnam(startLog);

   if (stat(pid, &fi) == 0)
   {
      sprintf(startProgram, "%s/restart-admin > %s 2>&1", sroot, startLog);
   }
   else
   {
      sprintf(startProgram, "%s/start-admin > %s 2>&1", sroot, startLog);
   }

   ret = system(startProgram);

   if (ret)
   {
      printf("Can't start Admin server [%s] (error: %s)", startProgram, system_errmsg());
      installLog->logMessage(WARN, "Admin", "Can't start Admin server [%s] (error: %s)", startProgram, system_errmsg());
   }
   else
   {
      FILE *fp = fopen (startLog, "r");
      char buf[BIG_BUF];

      if (fp)
      {
         while (fgets(buf, BIG_BUF, fp))
         {
            printf("%s\n",buf);
         }
         fclose(fp);
         unlink(startLog);
      }
   }
}

/*
void
walk_dir_and_chown(const char *dirName, uid_t my_uid, uid_t admin_uid, uid_t tuid, gid_t tgid)
{
   if (chdir(dirName) == 0)
   {
      struct dirent *entry = NULL;
      DIR *dir;
      struct stat fi;

      dir = opendir(".");
      while ((entry = readdir(dir)) != NULL)
      {
         if (strcmp(entry->d_name, ".") && strcmp(entry->d_name, "..") &&
                lstat(entry->d_name, &fi) == 0)
         {
            if (S_ISDIR(fi.st_mode))
            {
               walk_dir_and_chown(entry->d_name,
                   my_uid, admin_uid, tuid, tgid);
            }

            if ((my_uid == fi.st_uid || admin_uid == fi.st_uid)
                    && tuid != fi.st_uid)
            {
                lchown(entry->d_name, tuid, tgid);
            }
         }
      }

      closedir(dir);

      if (dirName[0] != '/')
      {
         chdir("..");
      }
   }
}
*/

void
admin_chown(const char *sroot, uid_t tuid, const char *tuser, gid_t tgid, const char *tgrp)
{
   /*
    * Changing ownership is possible only if I am installing as
    * root and the directories/files are owned by me or
    * the original admin person
    */
   struct dirent *entry = NULL;
   DIR *dir;
   struct stat fi;
   char tmp[BIG_BUF];
   uid_t my_uid;
   uid_t admin_uid;

   /*
    * First, check to see who currently is the admin user
    */
   sprintf(tmp, "%s/admin-serv", sroot);
   stat(tmp, &fi);

   admin_uid = fi.st_uid;
   my_uid    = geteuid();

   chown(sroot, tuid, tgid);

   dir = opendir(sroot);
   while ((entry = readdir(dir)) != NULL)
   {
      if (strcmp(entry->d_name, ".") && strcmp(entry->d_name, "..") &&
           lstat(entry->d_name, &fi) == 0)
      {
         if (S_ISDIR(fi.st_mode))
         {
            if ((my_uid == fi.st_uid || admin_uid == fi.st_uid) &&
                 tuid != fi.st_uid)
            {
               sprintf(tmp, CHOWN_CMD, tuser, sroot);
               system(tmp);
               sprintf(tmp, CHGRP_CMD, tgrp, sroot);
               system(tmp);
            }
         }
      }
   }
}

void
chown_file(const char *user, const char *group, const char *file)
{
    /* only do this if we are root */
    uid_t my_uid = geteuid();

    if (my_uid == 0) {
        char tmp[BIG_BUF];

        snprintf(tmp, sizeof(tmp), CHOWN_CMD, user, file);
        tmp[sizeof(tmp)-1] = 0;
        system(tmp);
        snprintf(tmp, sizeof(tmp), CHGRP_CMD, group, file);
        tmp[sizeof(tmp)-1] = 0;
        system(tmp);
    }
}    

void
create_admserv(char *hn, const char *sroot)
{
   char httpuser[SML_BUF], httppw[SML_BUF];
   char hosts[MED_BUF], addrs[MED_BUF];
   char *hexp, *aexp;
   char dir[BIG_BUF], tstr[BIG_BUF];
   char adminUser[SML_BUF];
   char SSuser[SML_BUF];
   char SSgroup[SML_BUF];
   char ldapURL[MED_BUF];
   char apacheRoot[BIG_BUF];
   char apacheDir[BIG_BUF];
   char apacheBin[BIG_BUF];
   char apacheVersion[BIG_BUF];
   struct stat st;
   NSString  hostip = InstUtil::getHostIpAddress(hn);
   char serverAddress[SML_BUF];
   struct stat statbuf;
   FILE *f;
   int port;
   const char *s;
   char *ss;
   NSString tmp;
   const char *adminBrand = "Fedora";

   chdir(sroot);
   
   if (NULL == adminInstallInfo) {
      fatal_error("Admin Info is not set or invalid.", "", "");
   }

   // Get admin port from installInfo
   s = adminInstallInfo->get("Port");
   if (NULL == s) {
      fatal_error("InstallInfo: Admin Port \"Port\" is missing.", "", "");
   }
   port = atoi(s);

   s = adminInstallInfo->get("SysUser");
   if (NULL == s) {
      fatal_error("InstallInfo: Admin System User \"SysUser\" is missing.", "", "");
   }
   strcpy(adminUser, s);

   s = adminInstallInfo->get("ApacheDir");
   if (NULL == s) {
      fatal_error("InstallInfo: Apache Directory \"ApacheDir\" is missing.", "", "");
   }
   strcpy(apacheDir, s);

   s = adminInstallInfo->get("ApacheRoot");
   if (!s) {
	   NSString ar = findApacheRoot(apacheDir);
	   strcpy(apacheRoot, (const char *)ar);
   } else {
	   strcpy(apacheRoot, s);
   }

   snprintf(apacheBin, sizeof(apacheBin), "%s/httpd.worker", apacheDir);
   if (stat(apacheBin, &st) != 0) {
      snprintf(apacheBin, sizeof(apacheBin), "%s/httpd", apacheDir);
      if (stat(apacheBin, &st) != 0) {
         fatal_error("Can't find Apache in %s", apacheDir, NULL);
      }
   }

   apacheVersion[0] = '\0';
   tmp = findApacheVersion(apacheDir);
   if ((const char*)tmp) {
       strcpy(apacheVersion, (const char *)tmp);
   }

   s = installInfo->get(SS_USER_ID);
   if (NULL == s) {
      fatal_error("InstallInfo: Server User \"SuiteSpotUserID\" is missing.", "", "");
   }
   strcpy(SSuser, s);

   s = installInfo->get(SS_GROUP);
   if (NULL == s) {
      fatal_error("InstallInfo: Server Group \"SuiteSpotGroup\" is missing.", "", "");
   }
   strcpy(SSgroup,s);

   s = adminInstallInfo->get("ServerIpAddress");
   if (s == NULL || s[0] == ' ')
      strcpy(serverAddress, "\0");
   else
      strcpy(serverAddress, s);

   char *p;

   p = strdup(installInfo->get(CONFIG_LDAP_URL));
   if (NULL == p) {
      fatal_error("Failed to duplicate %s.", installInfo->get(CONFIG_LDAP_URL), "");
   }

   char *cp;

   cp = escapeLDAPurl(p);

   if (cp)
   {
      strcpy(ldapURL, cp);
      free(p);
   }

   s = adminInstallInfo->get("ServerAdminID");
   if (s)
   {
      strcpy(httpuser, s);
   }
   else
   {
#if 0
      installLog->logMessage(WARN,"Admin","ServerAdminID not specified. Default to admin.");
#endif
      strcpy(httpuser, DEFAULT_ADMINUSER);
   }
   s = adminInstallInfo->get("ServerAdminPwd");

   if (s)
   {
      strcpy(httppw, s);
   }
   else
   {
#if 0
      installLog->logMessage(WARN,"Admin","ServerAdminPwd not specified. Use default.");
#endif
      strcpy(httppw, DEFAULT_ADMINPWD);
   }

#ifdef ORIGINAL_HOST_IP_QUERIES
   // PVO
   s = installInfo->get("HostIpAddress");
   strcpy(hosts,s);
#else
   strcpy(hosts, "*");
   if (strchr(hn, '.'))
   {
      strcat(hosts, strchr(hn, '.'));
   }

/*
 * PVO - temporary solution for nspr21 problem
   strcpy(addrs, "*");
*/
   if (hostip.data())
   {
      strcpy(addrs, hostip.data());
   }
   else
   {
      strcpy(addrs, "*");
   }

#endif

   if (installMode == Interactive)
   {
      printf("Your parameters are now entered into the Administration Server\n"
	     "database, and the Administration Server will be started.\n\n");
   }
   else
   {
      installLog->logMessage(INFO, "Admin", "Writing configuration data to %s/admin-serv/config", sroot);
   }

/* create alias subdirectory and cert.log file */
   sprintf(dir, "%s%calias", sroot, PATH_DELIM);
   if ((mkdir(dir, 0750) == -1) && (errno != EEXIST))
   {
      fatal_error("Can't create directory %s (error: %s)", dir, system_errmsg());
   }

   logUninstallInfo(sroot, "admin", "admin", dir);

   /* make sure the config directory is owned by the ssuser */
   sprintf(tstr, "%s/admin-serv/config", sroot);
   chown_file(SSuser, SSgroup, tstr);

   /*
    * Write new admpw.
    */
/* MLM - this should actually do more httpconf type of stuff
 *       (maybe an admconf.c?) */
   sprintf(tstr, "%s/admin-serv/config/admpw", sroot);
   if (!(f = fopen(tstr, "w")))
   {
      fatal_error("Can't write to %s (error: %s)", tstr, system_errmsg());
   }

   char *pw = sha1_pw_enc (httppw);
   fprintf(f, "%s:{SHA}%s", httpuser, pw);
   fclose(f);
   chmod(tstr, S_IRUSR | S_IWUSR);
   chown_file(SSuser, SSgroup, tstr);
   logUninstallInfo(sroot, "admin", "admin", tstr);

   /* create admin-serv/modules directory */
   sprintf(dir, "%s%cadmin-serv%cmodules", sroot, PATH_DELIM, PATH_DELIM);
   if ((mkdir(dir, 0750) == -1) && (errno != EEXIST))
   {
      fatal_error("Can't create directory %s (error: %s)", dir, system_errmsg());
   }

   logUninstallInfo(sroot, "admin", "admin", dir);

#ifdef SHOW_CONFIG_FILES
   if (installMode != Silent)
      printf("Wrote %s\n", tstr);
#endif

   /*
    * Write new LDAP local DB  default files
    */
   sprintf(tstr, "%s/%s", sroot, DEFAULT_LDAPSWITCH);
   if (!(f = fopen(tstr, "w")))
   {
      fatal_error("Can't write to %s (error: %s)", tstr, system_errmsg());
   }
   if (ldapURL[0] == '\0')
   {				/* No  external LDAP */
      fatal_error("No LDAP server defined - exiting", NULL, NULL);
   }
   else
   {				/* Use external LDAP */
      fprintf(f, "directory default %s\n", ldapURL);
   }
   fclose(f);

   logUninstallInfo(sroot, "admin", "admin", tstr);
   chown_file(SSuser, SSgroup, tstr);

   sprintf(tstr, "chmod 640 %s/%s", sroot, DEFAULT_LDAPSWITCH);
   system(tstr);

   hexp = make_shexp(hosts);
   aexp = make_shexp(addrs);

   sprintf(tstr, "%s/%s", sroot, DEFAULT_SSUSERCONF);
   if (!(f = fopen(tstr, "w")))
      fatal_error("Can't write to %s (error: %s)", tstr,
		  system_errmsg());
   fprintf(f, "SuiteSpotUser\t%s\n", SSuser);
   fprintf(f, "SuiteSpotGroup\t%s\n", SSgroup);
   fclose(f);

   logUninstallInfo(sroot, "admin", "admin", tstr);

   if (!geteuid())
   {
      /* Current pw ent, target pw ent */
      struct passwd *cpw;
      struct group  *tgp;
      uid_t tuid;
      gid_t tgid;

      setpwent();
      if (!(cpw = getpwnam(adminUser)))
      {
	 fatal_error("Can't get password entry for user %s.", adminUser, NULL);
      }
      tuid = cpw->pw_uid;
      tgid = cpw->pw_gid;

      if (!(tgp = getgrgid(tgid)))
      {
	 fatal_error("Can't get group entry for target group.", NULL, NULL);
      }
      endpwent();

      if (installMode == Interactive)
         printf("Changing ownership to admin user %s...\n", adminUser);
      admin_chown(sroot, tuid, adminUser, tgid, tgp->gr_name);

#ifdef HPUX
      /*
      setresuid((uid_t) -1, tuid, (uid_t) -1);
      setresgid((gid_t) -1, tgid, (gid_t) -1);
      */
      setresuid(tuid, tuid, (uid_t) -1);
      setresgid(tgid, tgid, (gid_t) -1);
#else
      seteuid(tuid);
      setegid(tgid);
#endif
   }


   {
      /*
       * This is ok because running user is either root or part of the
       * SSgroup
       */
      sprintf(tstr, SHARED_CHGRP_CMD, SSgroup, sroot);
      system(tstr);
      sprintf(tstr, SSCONF_CHGRP_CMD, SSgroup, sroot);
      system(tstr);
      sprintf(tstr, SECALIAS_CHGRP_CMD, SSgroup, sroot);
      system(tstr);
   }

   ////////////////////////////////////////////////////////////////
   // Now everything seems to be OKAY, install SIE for admin-server
   ////////////////////////////////////////////////////////////////

   LdapError ldapError;
   NSString  adminDomain;
   NSString  sieDN;
   NSString  isieDN;
   NSString  configDN;
   char *ldapStart = NULL;

   printf("Setting up Administration Server Instance...\n");
   installLog->logMessage(INFO, "Admin", "Setting up Administration Server Instance...\n");

   Ldap ldap (ldapError, ldapURL, installInfo->get(MC_ADMIN_ID),
                    installInfo->get(MC_ADMIN_PWD),
                    NULL, NULL); // NULL for now until figure out PVO

   if (ldapError != OKAY)
   {
      sprintf(dir, "%ld", ldapError.errorCode());
      fatal_error("ERROR: Ldap authentication failed (%s:%s)" , dir, ldapError.msg());
   }

   LdapEntry *sieEntry = new LdapEntry(&ldap);
   LdapEntry *appEntry = new LdapEntry(&ldap);
   LdapEntry *configEntry = new LdapEntry(&ldap);

   // Check if there's already an SIE for this Admin
   sprintf(tstr, "%s/admin-serv/config/adm.conf", sroot);
   NVPair admConf(tstr);

   adminDomain = installInfo->get(DS_ADMIN_DOMAIN);

   if (admConf.isEmpty() == False)
   {
      sieDN = admConf.get("sie");
      isieDN = admConf.get("isie");
      if (s=admConf.get("ldapStart"))
      {
         ldapStart = strdup(s);
      }
   }

   /*
    * Generate pwd for the SIE
    */

/*
   char password[9];

   srand(getpid() + time(NULL));
   for (int i = 0; i < 8 ; i++)
   {
      int charoff = rand() % 52;
      if (charoff < 26)
        password[i] = (char) charoff + 'A';
      else
        password[i] = (char) charoff - 26 + 'a';
   }

   password[8] = 0;

*/

   sprintf(tstr, "%s/%s/admin/admin.inf", sroot, SETUP_DIR);
   NVPair adminInf(tstr);
   NVPair *adminInfo = adminInf.getSection("admin");

   appEntry->addAttribute("objectclass","top");
   appEntry->addAttribute("objectclass","nsApplication");

   appEntry->setAttribute("cn", "Fedora Administration Server");

   appEntry->addAttribute("nsVendor", adminInfo->get(PKG_VENDOR));
   appEntry->addAttribute("description", adminInfo->get(PKG_DESCRIPTION));
   appEntry->addAttribute("nsProductName", adminInfo->get(PKG_NAME));
   appEntry->addAttribute("nsNickname", adminInfo->get(PKG_NICKNAME));
   appEntry->addAttribute("nsProductVersion", adminInfo->get(PKG_VERSION));
   appEntry->addAttribute("nsBuildNumber", adminInfo->get(PKG_BUILDNUMBER));
   appEntry->addAttribute("nsRevisionNumber", adminInfo->get(PKG_REVISION));
   appEntry->addAttribute("nsInstalledLocation", sroot);

   adminInfo->stampTime();
   appEntry->addAttribute("installationTimeStamp", adminInfo->timeStamp());
   appEntry->addAttribute("nsExpirationDate", adminInfo->get(PKG_EXPIRY));
   appEntry->addAttribute("nsBuildSecurity", adminInfo->get(PKG_SECURITY));
   appEntry->addAttribute("nsServerMigrationClassName", "com.netscape.management.admserv.AdminServerProduct@fedora-admserv-1.0.jar");

   sieEntry->clear();

 
    /*
     * set password expiration to some later date so 
     * admin server will not stop working due to password expiration
     * see bug#343369 - http://scopus/bugsplat/show_bug.cgi?id=343369
     * 
     * passwordExpirationTime is a hidden attribute.
     * For some reason if we don't add this attribute to the top of the
     * attribute list we will end up with 2 passwordExpirationTime 
     * attribute...
     * leave it here don't move into setupCreateSIE.
     *
     */
   sieEntry->setAttribute("passwordExpirationTime", EPOCH_32BIT_END_TIME);

   sieEntry->addAttribute("objectclass", "top");
   sieEntry->addAttribute("objectclass", "netscapeServer");
   sieEntry->addAttribute("objectclass", "nsAdminServer");
   sieEntry->addAttribute("objectclass", "nsResourceRef");

   /* Form the CN for Admin Server, e.g. admin-serv-lupine */
   if (sieDN.data() == NULL)
   {
      sprintf(tstr, "%s-%s", "admin-serv", hn);
      ss = strchr(tstr, '.');
      if (ss) *ss = 0;
   }
   else
   {
      // Get the CN part
      char **tmpDN = ldap_explode_dn(sieDN.data(), 1);
      strcpy(tstr, tmpDN[0]);
      ldap_value_free(tmpDN);
   }

   sieEntry->setAttribute("cn", tstr);
   sieEntry->addAttribute("nsServerId", "admin-serv");
   sieEntry->addAttribute("userPassword", httppw);
   sieEntry->addAttribute("serverHostName", hn);
   sieEntry->addAttribute("serverRoot", sroot);
   sieEntry->addAttribute("serverProductName", "Administration Server");

   ldapError = createSIE(sieEntry, appEntry, hn, sroot, adminDomain, adminBrand);

   sieDN = sieEntry->entryDN();
   isieDN = appEntry->entryDN();

   if (ldapError != OKAY)
   {
      sprintf(dir, "%d", ldapError.errorCode());
      fatal_error("ERROR: cannot create server instance (%s:%s)", dir, ldapError.msg()); // PVO
   }


   NSString configTaskDN, opTaskDN;

   // Write configuration parameters (see ns-admin-schema.conf)
   configDN = NSString("cn=configuration") + LDAP_PATHSEP + sieDN;

   sprintf(dir, "%d", port);

   const char *userDirRef = strchr(isieDN.data(), ',') + 1;
   while (*userDirRef == ' ') ++userDirRef;

   if (configEntry->retrieve(configDN) != OKAY)
   {
      configEntry->addAttribute("objectclass", "nsConfig");
      configEntry->addAttribute("objectclass", "nsAdminConfig");
      configEntry->addAttribute("objectclass", "nsAdminObject");
      configEntry->addAttribute ("cn", "Configuration");

      configEntry->addAttribute ("nsServerPort", dir);
      configEntry->addAttribute ("nsSuiteSpotUser", adminUser);
      configEntry->addAttribute ("nsServerAddress", serverAddress);

      configEntry->addAttribute ("nsAdminEnableEnduser", "on");
      configEntry->addAttribute ("nsAdminEnableDSGW", "on");

      /*
       * Mandatory fields here
       */
      configEntry->addAttribute("objectclass", "nsDirectoryInfo");
      configEntry->setAttribute("nsDirectoryInfoRef", userDirRef);

      configEntry->addAttribute ("nsAdminUsers", "admin-serv/config/admpw");
      configEntry->addAttribute ("nsErrorLog", "admin-serv/logs/error");
      configEntry->addAttribute ("nsPidLog",   "admin-serv/logs/pid");
      configEntry->addAttribute ("nsAccessLog","admin-serv/logs/access");
      configEntry->addAttribute ("nsAdminCacheLifetime", "600");

      if (hosts[0])
         configEntry->addAttribute ("nsAdminAccessHosts", hexp);
      //if (addrs[0])
      //   configEntry->addAttribute ("nsAdminAccessAddresses", aexp);
      configEntry->addAttribute ("nsAdminAccessAddresses", "*");
      configEntry->addAttribute ("nsAdminOneAClDir", "adminacl");

      sprintf(tstr, DEFAULT_ADMIN_CONFIG_ACI, userDirRef);
      configEntry->addAttribute("aci",tstr);

      /*
       * Don't set this value - absent is default
       * configEntry->addAttribute ("nsAdminEnableDsgw",
       */

      configEntry->addAttribute ("nsDefaultAcceptLanguage", adminInfo->get(PKG_DEFAULT_LANGUAGE));

      sprintf(tstr, "com.netscape.management.admserv.AdminServer@fedora-admserv-1.0.jar@%s", sieDN.data());
      configEntry->addAttribute ("nsClassname", tstr);

      // Set the appropriate ACI on for this branch
      createACIForConfigEntry(configEntry, sieDN);

      ldapError = configEntry->insert(configDN);
   }
   else
   {
      const char *s;
      sprintf(tstr, "com.netscape.management.admserv.AdminServer@fedora-admserv-1.0.jar@%s", sieDN.data());
      configEntry->setAttribute ("nsClassname", tstr);
      configEntry->setAttribute ("nsServerPort", dir);
      configEntry->setAttribute ("nsSuiteSpotUser", adminUser);
      
      /*
       * in custom mode, update IP address in DS
       */
      if ((s = installInfo->get("InstallType")) && (strcasecmp(s, "3") == 0))
         configEntry->setAttribute ("nsServerAddress", serverAddress);

      ldapError = configEntry->replace(configDN);
   }

   if (ldapError != OKAY)
   {
      sprintf(dir, "%d", ldapError.errorCode());
      fatal_error("ERROR: cannot write SIE configuration (%s:%s)", dir, ldapError.msg());
   }

   /*
    *  Write Tasks node
    */

   configTasks(&ldap, sroot, sieDN);

   // Write global configuration paramters

   sprintf(tstr, "%s/shared/config/ds.conf", sroot);
   if (!(f = fopen(tstr,"w")))
   {
      printf("Warning - Can't write to %s (error: %s)", tstr, system_errmsg());
   }
   else
   {
      fprintf(f, "%s:%s\n", DS_ADMIN_DOMAIN, adminDomain.data());
      logUninstallInfo(sroot, "admin", "admin", tstr);
   }

   char *adminDomainDN;
   adminDomainDN = formAdminDomainDN(adminDomain);
   configGlobalParameters(&ldap, sroot, adminDomainDN);
   nsSetupFree (adminDomainDN);

   /*
    ***************************************************************
    * Fix ACIs under o=netscapeRoot
    ***************************************************************
    */
   /*ldapError = repair_group_expansion_aci(&ldap);
   if (ldapError != OKAY)
   {
      sprintf(dir, "%d", ldapError.errorCode());
      fatal_error("ERROR: cannot modify root ACIs (%s:%s)", dir, ldapError.msg()); 
   }*/

   /*ldapError = set_locate_admserv_aci(&ldap);
   if (ldapError != OKAY)
   {
      sprintf(dir, "%d", ldapError.errorCode());
      fatal_error("ERROR: cannot modify root ACIs (%s:%s)", dir, ldapError.msg()); 
   }*/

   /*
    ***************************************************************
    * Write new adm.conf (new for Admin40)
    ***************************************************************
    */

   LDAPURLDesc *ludpp = NULL;
   ldap_url_parse((char *) installInfo->get(CONFIG_LDAP_URL), &ludpp);

/* 

   fprintf(f, "ldapHost: %s\n", ludpp->lud_host);
   fprintf(f, "ldapPort: %d\n", ludpp->lud_port);
   fprintf(f, "sie: %s\n", sieDN.data());
   fprintf(f, "userdn: %s\n", ldap.rebindDN());
   fprintf(f, "isie: %s\n", isieDN.data());
   fprintf(f, "port: %d\n", port);
   fclose(f);

*/

   admConf.set("ldapHost",  ludpp->lud_host);
   admConf.set("ldapPort",  (unsigned long)ludpp->lud_port);
   admConf.set("sie", sieDN.data());
   admConf.set("userdn", ldap.rebindDN());
   admConf.set("isie", isieDN.data());
   admConf.set("port",  (unsigned long)port);

   if (ldapStart == NULL)
   {
      const char *isConfigDS = NULL;
      const char *dsServerID = NULL;

      dsInfo = installInfo->getSection("slapd");
      if (dsInfo != NULL)
      {
          isConfigDS = dsInfo->get("SlapdConfigForMC");
          dsServerID = dsInfo->get("ServerIdentifier");
          if (isConfigDS && dsServerID && !strcasecmp(isConfigDS, "Yes"))
          {
             sprintf(tstr, "slapd-%s/start-slapd", dsServerID);
             ldapStart = strdup(tstr);
          }
      }
   }

   if (ldapStart)
   {
      admConf.set("ldapStart", ldapStart);
   }

   sprintf(tstr, "%s/admin-serv/config/adm.conf", sroot);

   admConf.setFormat(2);
   admConf.write(tstr);
   chmod(tstr, S_IRUSR | S_IWUSR);
   chown_file(SSuser, SSgroup, tstr);
   logUninstallInfo(sroot, "admin", "admin", tstr);

   // Mark to remove the following guys which are
   // created later on also
   sprintf(tstr, "%s/admin-serv/config/local.conf", sroot);
   logUninstallInfo(sroot, "admin", "admin", tstr);
   sprintf(tstr, "%s/admin-serv/logs/access", sroot);
   logUninstallInfo(sroot, "admin", "admin", tstr);
   sprintf(tstr, "%s/admin-serv/logs/error", sroot);
   logUninstallInfo(sroot, "admin", "admin", tstr);
   sprintf(tstr, "%s/admin-serv/logs/admpw.log", sroot);
   logUninstallInfo(sroot, "admin", "admin", tstr);
   sprintf(tstr, "%s/admin-serv/logs/dsconfig.log", sroot);
   logUninstallInfo(sroot, "admin", "admin", tstr);

   /* PVO remember to set access permission here */

   char **httparray = (char **)malloc(5*sizeof(char *)); /* 2 tokens, 2 replacements, NULL terminated */
   char **admservarray = (char **)malloc(3*sizeof(char *)); /* 1 token, 1 replacement, NULL terminated */
   char **consolearray = (char **)malloc(11*sizeof(char *)); /* 5 tokens, 5 replacements, NULL terminated */
   char **nssarray = (char **)malloc(3*sizeof(char *)); /* 1 token, 1 replacement, NULL terminated */
   char **startarray = (char **)malloc(5*sizeof(char *)); /* 2 tokens, 2 replacements, NULL terminated */
   char **stoparray = (char **)malloc(3*sizeof(char *)); /* 1 token, 1 replacement, NULL terminated */
   char **restartarray = (char **)malloc(3*sizeof(char *)); /* 1 token, 1 replacement, NULL terminated */

   char src[BIG_BUF];
   char instance[BIG_BUF];

   snprintf(instance, sizeof(instance), "%s-%s", "admin-serv", hn);
   ss = strchr(instance, '.');
   if (ss) *ss = 0;

   httparray[0] = "%%%sroot%%%";
   httparray[1] = (char *)sroot;
   httparray[2] = "%%%module_dir%%%";
   httparray[3] = (char *)apacheRoot;
   httparray[4] = NULL;

   admservarray[0] = "%%%sroot%%%";
   admservarray[1] = (char *)sroot;
   admservarray[2] = NULL;

   nssarray[0] = "%%%sroot%%%";
   nssarray[1] = (char *)sroot;
   nssarray[2] = NULL;

   consolearray[0] = "%%%sroot%%%";
   consolearray[1] = (char *)sroot;
   consolearray[2] = "%%%httpduser%%%";
   consolearray[3] = (char *)SSuser;
   consolearray[4] = "%%%httpdgroup%%%";
   consolearray[5] = (char *)SSgroup;
   consolearray[6] = "%%%httpport%%%";
   consolearray[7] = (char *)dir; /* sprint'd earlier */
   consolearray[8] = "%%%instancename%%%";
   consolearray[9] = (char *)instance;
   consolearray[10] = NULL;

   startarray[0] = "%%%sroot%%%";
   startarray[1] = (char *)sroot;
   startarray[2] = "%%%httpd%%%";
   startarray[3] = (char *)apacheBin;
   startarray[4] = NULL;

   stoparray[0] = "%%%sroot%%%";
   stoparray[1] = (char *)sroot;
   stoparray[2] = NULL;

   restartarray[0] = "%%%sroot%%%";
   restartarray[1] = (char *)sroot;
   restartarray[2] = NULL;

   if (0 == strncmp(apacheVersion, "2.2", 3)) {
       snprintf(src, sizeof(tstr), "%s/shared/config/template/httpd-2.2.conf.tmpl", sroot);
   } else {
       snprintf(src, sizeof(tstr), "%s/shared/config/template/httpd.conf.tmpl", sroot);
   }
   snprintf(tstr, sizeof(tstr), "%s/admin-serv/config/httpd.conf", sroot);
   replaceTokensInFile(src, tstr, httparray);
   chmod(tstr, S_IRUSR | S_IWUSR);
   logUninstallInfo(sroot, "admin", "admin", tstr);

   snprintf(tstr, sizeof(tstr), "%s/admin-serv/config/admserv.conf", sroot);
   snprintf(src, sizeof(tstr), "%s/shared/config/template/admserv.conf.tmpl", sroot);
   replaceTokensInFile(src, tstr, admservarray);
   chmod(tstr, S_IRUSR | S_IWUSR);
   logUninstallInfo(sroot, "admin", "admin", tstr);

   snprintf(tstr, sizeof(tstr), "%s/admin-serv/config/nss.conf", sroot);
   snprintf(src, sizeof(tstr), "%s/shared/config/template/nss.conf.tmpl", sroot);
   replaceTokensInFile(src, tstr, nssarray);
   chmod(tstr, S_IRUSR | S_IWUSR);
   chown_file(SSuser, SSgroup, tstr);
   logUninstallInfo(sroot, "admin", "admin", tstr);

   snprintf(tstr, sizeof(tstr), "%s/admin-serv/config/console.conf", sroot);
   /*
     all of the dynamic information goes into console.conf, so we should
     not overwrite that information during an upgrade */
   if (stat(tstr, &st)) {
      snprintf(src, sizeof(tstr), "%s/shared/config/template/console.conf.tmpl", sroot);
      replaceTokensInFile(src, tstr, consolearray);
      chmod(tstr, S_IRUSR | S_IWUSR);
      chown_file(SSuser, SSgroup, tstr);
      logUninstallInfo(sroot, "admin", "admin", tstr);
   }

   snprintf(tstr, sizeof(tstr), "%s/start-admin", sroot);
   snprintf(src, sizeof(tstr), "%s/shared/config/template/start-admin.tmpl", sroot);
   replaceTokensInFile(src, tstr, startarray);
   chmod(tstr, 0755);
   logUninstallInfo(sroot, "admin", "admin", tstr);

   snprintf(tstr, sizeof(tstr), "%s/stop-admin", sroot);
   snprintf(src, sizeof(tstr), "%s/shared/config/template/stop-admin.tmpl", sroot);
   replaceTokensInFile(src, tstr, stoparray);
   chmod(tstr, 0755);
   logUninstallInfo(sroot, "admin", "admin", tstr);

   snprintf(tstr, sizeof(tstr), "%s/restart-admin", sroot);
   snprintf(src, sizeof(tstr), "%s/shared/config/template/restart-admin.tmpl", sroot);
   replaceTokensInFile(src, tstr, restartarray);
   chmod(tstr, 0755);
   logUninstallInfo(sroot, "admin", "admin", tstr);
   
   sprintf(tstr, "%s/java/jars", sroot);
   chmod(tstr, 0755);

   sprintf(tstr, "%s/bin/admin/admconfig", sroot);
   chmod(tstr, 0755);

   sprintf(tstr, "%s/uninstall", sroot);
   chmod(tstr, 0750);

   {
      // Generating local.conf
      int error_code = 0;
      snprintf(tstr, sizeof(tstr), "%s/admin-serv/config", sroot);
      PsetHndl pset = psetCreate("admin-serv", (char *)tstr,
                                 (char *)ldap.rebindDN(), httppw,
                                 &error_code);
   }
}


int
update_schema(const char *sroot)
{
  char schema_file[BIG_BUF];
  
  sprintf(schema_file, "%s/setup/admin/ns-common-schema.conf", sroot);
  return setupInsertPluginSchemaEntries(installInfo->get(CONFIG_LDAP_URL),
					installInfo->get(MC_ADMIN_ID),
					installInfo->get(MC_ADMIN_PWD),
					schema_file);
}

/*
 * 359259: Add ACIs under o=netscapeRoot to allow regular, non-admin users to
 * locate admin servers. This is required for downloads of jars referenced from
 *  Users and Groups configuration under ou=Global Preferences.
 */
static int set_locate_admserv_aci(Ldap *ldap) {
    LDAPMod *mods[2];
    LdapErrorCode ldapError;
    LDAPMod mod;
    char *aciVal[2];

    // Allow read access to host, port, address and security attributes for SIEs
    aciVal[0] = (char*)LOCATE_ADMSERV_ACI;
    aciVal[1] = NULL;

    mod.mod_op = LDAP_MOD_ADD;
    mod.mod_type = "aci";
    mod.mod_values = aciVal;

    mods[0] = &mod;
    mods[1] = NULL;

    ldapError = ldap->modify("o=netscapeRoot", mods);

    if (ldapError != LDAP_SUCCESS &&
        ldapError != LDAP_TYPE_OR_VALUE_EXISTS) {
        fprintf(stderr, 
           "Could not modify ACI for o=netscapeRoot, \"%s\"", ldap_err2string(ldapError));
        return ldapError;
    }

    return OKAY;
}

/* 
 * 337487: Fix "Enable Group Expansion" ACI under o=netscapeRoot. The ACI
 * is defined as <groupdnattr="ldap:///o=NetscapeRoot?uniquemember?sub">.
 * This is instructing that all groups will be located under the
 * o=netscapeRoot suffix, which is true only for administrators. Set
 * <groupdnattr="uniquemember"> which allows a group to be under any suffix.
 */
static int repair_group_expansion_aci(Ldap *ldap) {
   LdapError ldapError;
   LdapEntry *root = new LdapEntry(ldap);
   char ** aciVals, *aci;

    ldapError = root->retrieve("o=netscapeRoot");
    if (ldapError != OKAY) {
        fprintf(stderr, "Cannot read o=netscapeRoot err=%d\n", (int)ldapError);
        return ldapError;
    }
    
    aciVals = root->getAttributes("aci");
    while (aci=*aciVals++) {

        if (strstr(aci, "Enable Group Expansion") &&
            strstr(aci, "groupdnattr=\"ldap:///o=NetscapeRoot?uniquemember?sub\"")) {
        
            LDAPMod *mods[3];
            LDAPMod mod1, mod2;
            char *prevVal[2], *newVal[2];

            prevVal[0] = aci;
            prevVal[1] = NULL;

            newVal[0]  = (char*)GROUP_EXPANSION_ACI;
            newVal[1]  = NULL;

            mod1.mod_op = LDAP_MOD_DELETE;
            mod1.mod_type = "aci";
            mod1.mod_values = prevVal;

            mod2.mod_op = LDAP_MOD_ADD;
            mod2.mod_type = "aci";
            mod2.mod_values = newVal;

            mods[0] = &mod1;
            mods[1] = &mod2;
            mods[2] = NULL;

            ldapError = ldap->modify("o=netscapeRoot", mods);

            if (ldapError != OKAY) {
                fprintf(stderr, "Failed to modify o=netscapeRoot ACI ldapError=%s\n", ldap_err2string(ldapError));
                return ldapError;
           }
       }
    }

    return OKAY;
}

/*
  Substitutes an occurance of token with replacement in the given string.
  str should point at the beginning of token e.g. use like this:
  if (str = strstr(line, token)) {
     replaceToken(str, token, replacement);
  }

  str is modified in place, so it should have enough room for the replacement
*/
static void
replaceTokenInStr(char *str, const char *token, const char *replacement)
{
    int offset;
    int tlen = strlen(token); /* length of the token string */
    int rlen = strlen(replacement); /* length of the replacement string */
    char *limit = str + (tlen < rlen ? tlen : rlen); /* where token or repl ends, shortest */
    const char *rp = replacement; /* for iteration */

    /* up to limit, it is safe to just replace the token with the replacement */
    while (str < limit) {
        *str++ = *rp++;
    }

    /* at this point, we have 1 of 3 states:
       tlen == rlen - the token is exactly as long as the replacement
           we are finished
       tlen > rlen - the token is longer than the replacement - we need
           to copy back str to "fill in" the string
       tlen < rlen - the replacement is longer than the token - we need
           to copy forward str to make room for the replacement
    */
    offset = rlen - tlen;
    if (offset > 0) {
        /* first move the pointer to the end of the string */
        while (*str) {
            str++;
        }
        /* str now points to the null terminator - we need to copy that too */
        /* next, move the chars to make room for the replacement */
        while (str >= limit) {
            *(str + offset) = *str;
            str--;
        }
        /* finally, finish writing the replacement */
        while (*rp) {
            *limit++ = *rp++;
        }
    } else if (offset < 0) {
        /* move str back to fill in space */
        while (1) {
            /* offset is negative, so - a - == a + */
            *str = *(str - offset);
            if (!*str) { /* copied null - exit */
                break;
            }
            str++;
        }
    } /* else offset is 0 - done */

    return;
}

/*
  copy sourcefile to destfile, replacing the tokens in source file with the
  replacement values in destfile.  The array ary is an array of strings of
  alternately tokens and their replacements e.g.
  %%%sroot%%%
  /opt/fedora-ds
  %%%port%%%
  3234
  %%%httpd%%%
  /usr/sbin/httpd.worker
  NULL

  ary must be NULL terminated
*/
static void
replaceTokensInFile(const char *sourcefile, const char *destfile, const char * const *ary)
{
    FILE *src, *dest;
    char line[2048];

    src = fopen (sourcefile, "r");
    if (src == NULL) {
       fatal_error("Cannot open %s", sourcefile, 0);
    }
    
    /* The output file */
    dest = fopen (destfile, "w");
    if (dest == NULL) {
        fatal_error("Cannot open %s for writing", destfile, 0);
    }

    while (fgets(line, sizeof(line), src)) {
        const char * const * argp = ary;
        for (; argp && *argp; argp += 2) {
            const char *token = *argp;
            const char *replacement = *(argp + 1);
            char *str;
            while (str = strstr(line, token)) {
                replaceTokenInStr(str, token, replacement);
            }
        }
        fputs(line, dest);
    }

    fclose(dest);
    fclose(src);
    
}
       
/* -----------------------         main            ------------------------ */

main(int argc, char *argv[])
{
   char hn[SML_BUF];
   const char *sroot;
   struct stat fi;
   int opt;

   while ((opt = getopt(argc, argv, "sl:f:r")) != -1)
   {
      switch (opt)
	 {
	    case 'r':
	        reConfig = True;
	        break;
	    case 's':
	       installMode = Silent;
	       break;
	    case 'l':
	       logFile = strdup(optarg);	/* Log file to use */
	       break;
	    case 'f':
	       infoFile = strdup(optarg);	/* Install script */
               installInfo = new InstallInfo(infoFile);
	       break;
	    default:
	       break;
	 }
   }

   /*
    * ns-update always expects the -f option.
    */
   if (installMode == Silent)
   {
      installLog = new InstallLog(logFile);
      installLog->logMessage(START, "Admin", "Administration Server post-configuration");
   }

   if (infoFile == NULL || installInfo == NULL ||
       installInfo->isEmpty() == True ||
       (adminInstallInfo = installInfo->getSection("admin")) == NULL)
   {
      installLog->logMessage(FATAL, "Admin", "Answer cache (admin section) not found or invalid.");
      printf("ERROR: Answer cache (admin section) not found or invalid.\n");
      exit(1);
   }

   if (installMode == Silent)
   {
      if (infoFile == NULL || (stat(infoFile, &fi) == -1))
      {
        installLog->logMessage(FATAL, "Admin", "Installation script not found");
        printf("ERROR:Installation script not found.\n");
        exit(1);
      }
      else
      {
        // PVO - why ???
        installInfo = new InstallInfo (infoFile);
      }

   }

   // Some additional checking in case
   if (installInfo->get(MACHINE_NAME) == NULL ||
       installInfo->get(SERVER_ROOT) == NULL ||
       installInfo->get(CONFIG_LDAP_URL) == NULL ||
       installInfo->get(MC_ADMIN_ID) == NULL ||
       installInfo->get(MC_ADMIN_PWD) == NULL ||
       installInfo->get(DS_ADMIN_DOMAIN)==NULL)
   {
      installLog->logMessage(FATAL, "Admin", "Missing data from answer cache");
      printf("ERROR: missing data from answer cache\n");
      exit(1);
   }

   // Add config DN (o=NetscapeRoot) to CONFIG_LDAP_URL
   installInfo->set(CONFIG_LDAP_URL, formConfigLdapURL(installInfo->get(CONFIG_LDAP_URL)));

   installInfo->toUTF8(CONFIG_LDAP_URL);
   installInfo->toUTF8(MC_ADMIN_ID);
   installInfo->toUTF8(DS_ADMIN_DOMAIN);

   installLog->logMessage(INFO, "Admin", "Configuring Administration server...");
   printf("Configuring Administration Server...\n");

   hn[0] = '\0';

   if (stat(DEFAULT_ADMINCONF, &fi) == 0)
   {
      // PVO - already configured
      // Probably should check DS instead
   }

   sroot = installInfo->get(SERVER_ROOT);
   if (NULL == sroot) {
      fatal_error("InstallInfo: Server Root \"ServerRoot\" is missing.", "", "");
   }

   /*
    * Get the full hostname.
    */
   const char *machine_name;
   machine_name = installInfo->get(MACHINE_NAME);
   if (NULL == machine_name) {
      fatal_error("InstallInfo: Hostname \"FullMachineName\" is missing.", "", "");
   }
   strcpy(hn, machine_name);

   cleanup_after_upgrade(sroot);

   create_admserv(hn, sroot);

#ifdef NEED_ADMIN_SERVER_IN_SERVERS_LIST
   addToServerList(sroot);
#endif

   start_admserv(sroot);

   installLog->logMessage(SUCCESS, "Admin", "Administration Server post-configuration");

   return 0;
}
